# Orca
#
# Copyright 2004-2009 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Provides a custom script for gcalctool."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2004-2009 Sun Microsystems Inc."
__license__   = "LGPL"

import orca.speech_generator as speech_generator
import pyatspi

class SpeechGenerator(speech_generator.SpeechGenerator):
    """Overrides _generateName to handle 'unspeakable' button labels
    displayed on the screen.
    """
    def __init__(self, script):
        speech_generator.SpeechGenerator.__init__(self, script)

    def _generateName(self, obj, **args):
        """Gives preference to the object name versus what is being
        displayed on the screen.  This helps accomodate the naming
        hints being given to us by gcalctool for it's mathematical
        operator buttons."""

        if not obj.getRole() in [pyatspi.ROLE_PUSH_BUTTON,
                                 pyatspi.ROLE_TOGGLE_BUTTON]:
            return speech_generator.SpeechGenerator._generateName(\
                self, obj)

        result = []
        acss = self.voice(speech_generator.DEFAULT)
        if obj.name:
            name = obj.name
        else:
            name = self._script.utilities.displayedText(obj)

        if name:
            result.append(name)
        elif obj.description:
            result.append(obj.description)
        if result:
            result.extend(acss)
        return result



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script for gnome-mud."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
__license__   = "LGPL"

import pyatspi

import orca.debug as debug
import orca.scripts.default as default
import orca.input_event as input_event
import orca.keybindings as keybindings
import orca.orca_state as orca_state
import orca.settings as settings
import orca.speech as speech

from orca.orca_i18n import _ # for gettext support

########################################################################
#                                                                      #
# Ring List. A fixed size circular list by Flavio Catalani             #
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/435902       #
#                                                                      #
########################################################################

class RingList:
    def __init__(self, length):
        self.__data__ = []
        self.__full__ = 0
        self.__max__ = length
        self.__cur__ = 0

    def append(self, x):
        if self.__full__ == 1:
            for i in range (0, self.__cur__ - 1):
                self.__data__[i] = self.__data__[i + 1]
            self.__data__[self.__cur__ - 1] = x
        else:
            self.__data__.append(x)
            self.__cur__ += 1
            if self.__cur__ == self.__max__:
                self.__full__ = 1

    def get(self):
        return self.__data__

    def remove(self):
        if (self.__cur__ > 0):
            del self.__data__[self.__cur__ - 1]
            self.__cur__ -= 1

    def size(self):
        return self.__cur__

    def maxsize(self):
        return self.__max__

    def __str__(self):
        return ''.join(self.__data__)


class Script(default.Script):

    MESSAGE_LIST_LENGTH = 10

    def __init__(self, app):
        """Creates a new script for the given application.
           This script tries to fix some accessibility problems found in
           the gnome-mud application, and also improves the user experience. 
           For more details see bug #
 

        Arguments:
        - app: the application to create a script for.
        """
        # Set the debug level for all the methods in this script.
        #
        self.debugLevel = debug.LEVEL_FINEST

        self.previousMessages = RingList(Script.MESSAGE_LIST_LENGTH)

        # Initially populate the cyclic list with empty strings
        i = 0
        while i < self.previousMessages.maxsize():
            self.previousMessages.append("")
            i += 1

        default.Script.__init__(self, app)


    def setupInputEventHandlers(self):
        debug.println(self.debugLevel, "gnome-mud.setupInputEventHandlers.")

        default.Script.setupInputEventHandlers(self)
        self.inputEventHandlers["readPreviousMessageHandler"] = \
            input_event.InputEventHandler(
                Script.readPreviousMessage,
                _('Read the latest n messages in the incoming messages text '
                  'area.'))

    def getAppKeyBindings(self):
        """Returns the application-specific keybindings for this script."""

        keyBindings = keybindings.KeyBindings()

        messageKeys = [ "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9"]
        for messagekey in messageKeys:
            keyBindings.add(
                keybindings.KeyBinding(
                    messagekey,
                    settings.defaultModifierMask,
                    settings.ORCA_MODIFIER_MASK,
                    self.inputEventHandlers["readPreviousMessageHandler"]))

        return keyBindings

    def readPreviousMessage(self, inputEvent):
        #This function speaks the latest n messages. Orca+F1 the latest one, 
        #Orca+F2 the latest two and so.

        debug.println(self.debugLevel, "gnome-mud.readPreviousMessage.")

        i = int(inputEvent.event_string[1:])
        messageNo = Script.MESSAGE_LIST_LENGTH - i
     
        text = ""
        messages = self.previousMessages.get()
        for i in range (messageNo, Script.MESSAGE_LIST_LENGTH):
            message = messages[i]
            text += message

        speech.speak(text)


    def onTextInserted(self, event):

        #Whenever a new text is inserted in the incoming message text area,
        #We want to speak and add it to the ringList structure only those lines
        #that contain some text and if the application is the current
        #locusOfFocus. 
        rolesList = [pyatspi.ROLE_TERMINAL,
                     pyatspi.ROLE_FILLER]
        if self.utilities.hasMatchingHierarchy(event.source, rolesList):
            if self.flatReviewContext:
                self.toggleFlatReviewMode()
            message = event.any_data
            if message and (not message.isspace()) and message != "\n":
                debug.println(debug.LEVEL_FINEST, \
                    message + " inserted in ringlist:")
                self.previousMessages.append(message)
                if event.source.getApplication() == \
                   orca_state.locusOfFocus.getApplication():
                    speech.speak(message)
        else:
            default.Script.onTextInserted(self, event)



# -*- coding: utf-8 -*-
# Orca
#
# Copyright 2004-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Provides getCharacterName that maps punctuation marks and other
individual characters into localized words."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
__license__   = "LGPL"

from .orca_i18n import _

# chnames is a dictionary where the keys represent a unicode character
# and the values represent the common term used for the character.
#
chnames = {}

# Translators: this is the spoken word for the space character
#
chnames[" "] = _("space")

# Translators: this is the spoken word for the newline character
#
chnames["\n"] = _("newline")

# Translators: this is the spoken word for the tab character
#
chnames["\t"] = _("tab")

# Translators: this is the spoken word for the character '!' (U+0021)
#
chnames["!"] = _("exclaim")

# Translators: this is the spoken word for the character '"' (U+0022)
#
chnames["\""] = _("quote")

# Translators: this is the spoken word for the character '#' (U+0023)
#
chnames["#"] = _("number")

# Translators: this is the spoken word for the character '$' (U+0024)
#
chnames["$"] = _("dollar")

# Translators: this is the spoken word for the character '%' (U+0025)
#
chnames["%"] = _("percent")

# Translators: this is the spoken word for the character '&' (U+0026)
#
chnames["&"] = _("and")

# Translators: this is the spoken word for the character ''' (U+0027)
#
chnames["'"] = _("apostrophe")

# Translators: this is the spoken word for the character '(' (U+0028)
#
chnames["("] = _("left paren")

# Translators: this is the spoken word for the character ')' (U+0029)
#
chnames[")"] = _("right paren")

# Translators: this is the spoken word for the character '*' (U+002a)
#
chnames["*"] = _("star")

# Translators: this is the spoken word for the character '+' (U+002b)
#
chnames["+"] = _("plus")

# Translators: this is the spoken word for the character ',' (U+002c)
#
chnames[","] = _("comma")

# Translators: this is the spoken word for the character '-' (U+002d)
#
chnames["-"] = _("dash")

# Translators: this is the spoken word for the character '.' (U+002e)
#
chnames["."] = _("dot")

# Translators: this is the spoken word for the character '/' (U+002f)
#
chnames["/"] = _("slash")

# Translators: this is the spoken word for the character ':' (U+003a)
#
chnames[":"] = _("colon")

# Translators: this is the spoken word for the character ';' (U+003b)
#
chnames[";"] = _("semicolon")

# Translators: this is the spoken word for the character '< ' (U+003c)
#
chnames["<"] = _("less")

# Translators: this is the spoken word for the character '=' (U+003d)
#
chnames["="] = _("equals")

# Translators: this is the spoken word for the character '> ' (U+003e)
#
chnames[">"] = _("greater")

# Translators: this is the spoken word for the character '?' (U+003f)
#
chnames["?"] = _("question")

# Translators: this is the spoken word for the character '@' (U+0040)
#
chnames["@"] = _("at")

# Translators: this is the spoken word for the character '[' (U+005b)
#
chnames["["] = _("left bracket")

# Translators: this is the spoken word for the character '\' (U+005c)
#
chnames["\\"] = _("backslash")

# Translators: this is the spoken word for the character ']' (U+005d)
#
chnames["]"] = _("right bracket")

# Translators: this is the spoken word for the character '^' (U+005e)
#
chnames["^"] = _("caret")

# Translators: this is the spoken word for the character '_' (U+005f)
#
chnames["_"] = _("underline")

# Translators: this is the spoken word for the character '`' (U+0060)
#
chnames["`"] = _("grave")

# Translators: this is the spoken word for the character '{' (U+007b)
#
chnames["{"] = _("left brace")

# Translators: this is the spoken word for the character '|' (U+007c)
#
chnames["|"] = _("vertical bar")

# Translators: this is the spoken word for the character '}' (U+007d)
#
chnames["}"] = _("right brace")

# Translators: this is the spoken word for the character '~' (U+007e)
#
chnames["~"] = _("tilde")

# Translators: this is the spoken character for the no break space
# character (e.g., "&nbsp;" in HTML -- U+00a0)
#
chnames['\u00a0'] = _("no break space")

# Translators: this is the spoken word for the character '¡' (U+00a1)
#
chnames['\u00a1'] = _("inverted exclamation point")

# Translators: this is the spoken word for the character '¢' (U+00a2)
#
chnames['\u00a2'] = _("cents")

# Translators: this is the spoken word for the character '£' (U+00a3)
#
chnames['\u00a3'] = _("pounds")

# Translators: this is the spoken word for the character '¤' (U+00a4)
#
chnames['\u00a4'] = _("currency sign")

# Translators: this is the spoken word for the character '¥' (U+00a5)
#
chnames['\u00a5'] = _("yen")

# Translators: this is the spoken word for the character '¦' (U+00a6)
#
chnames['\u00a6'] = _("broken bar")

# Translators: this is the spoken word for the character '§' (U+00a7)
#
chnames['\u00a7'] = _("section")

# Translators: this is the spoken word for the character '¨' (U+00a8)
#
chnames['\u00a8'] = _("umlaut")

# Translators: this is the spoken word for the character '©' (U+00a9)
#
chnames['\u00a9'] = _("copyright")

# Translators: this is the spoken word for the character 'ª' (U+00aa)
#
chnames['\u00aa'] = _("superscript a")

# Translators: this is the spoken word for the character '«' (U+00ab)
#
chnames['\u00ab'] = _("left double angle bracket")

# Translators: this is the spoken word for the character '¬' (U+00ac)
#
chnames['\u00ac'] = _("logical not")

# Translators: this is the spoken word for the character '­' (U+00ad)
#
chnames['\u00ad'] = _("soft hyphen")

# Translators: this is the spoken word for the character '®' (U+00ae)
#
chnames['\u00ae'] = _("registered")

# Translators: this is the spoken word for the character '¯' (U+00af)
#
chnames['\u00af'] = _("macron")

# Translators: this is the spoken word for the character '°' (U+00b0)
#
chnames['\u00b0'] = _("degrees")

# Translators: this is the spoken word for the character '±' (U+00b1)
#
chnames['\u00b1'] = _("plus or minus")

# Translators: this is the spoken word for the character '²' (U+00b2)
#
chnames['\u00b2'] = _("superscript 2")

# Translators: this is the spoken word for the character '³' (U+00b3)
#
chnames['\u00b3'] = _("superscript 3")

# Translators: this is the spoken word for the character '´' (U+00b4)
#
chnames['\u00b4'] = _("acute accent")

# Translators: this is the spoken word for the character 'µ' (U+00b5)
#
chnames['\u00b5'] = _("mu")

# Translators: this is the spoken word for the character '¶' (U+00b6)
#
chnames['\u00b6'] = _("paragraph marker")

# Translators: this is the spoken word for the character '·' (U+00b7)
#
chnames['\u00b7'] = _("middle dot")

# Translators: this is the spoken word for the character '¸' (U+00b8)
#
chnames['\u00b8'] = _("cedilla")

# Translators: this is the spoken word for the character '¹' (U+00b9)
#
chnames['\u00b9'] = _("superscript 1")

# Translators: this is the spoken word for the character 'º' (U+00ba)
#
chnames['\u00ba'] = _("ordinal")

# Translators: this is the spoken word for the character '»' (U+00bb)
#
chnames['\u00bb'] = _("right double angle bracket")

# Translators: this is the spoken word for the character '¼' (U+00bc)
#
chnames['\u00bc'] = _("one fourth")

# Translators: this is the spoken word for the character '½' (U+00bd)
#
chnames['\u00bd'] = _("one half")

# Translators: this is the spoken word for the character '¾' (U+00be)
#
chnames['\u00be'] = _("three fourths")

# Translators: this is the spoken word for the character '¿' (U+00bf)
#
chnames['\u00bf'] = _("inverted question mark")

# Translators: this is the spoken word for the character 'á' (U+00e1)
#
chnames['\u00e1'] = _("a acute")

# Translators: this is the spoken word for the character 'À' (U+00c0)
#
chnames['\u00c0'] = _("A GRAVE")

# Translators: this is the spoken word for the character 'Á' (U+00c1)
#
chnames['\u00c1'] = _("A ACUTE")

# Translators: this is the spoken word for the character 'Â' (U+00c2)
#
chnames['\u00c2'] = _("A CIRCUMFLEX")

# Translators: this is the spoken word for the character 'Ã' (U+00c3)
#
chnames['\u00c3'] = _("A TILDE")

# Translators: this is the spoken word for the character 'Ä' (U+00c4)
#
chnames['\u00c4'] = _("A UMLAUT")

# Translators: this is the spoken word for the character 'Å' (U+00c5)
#
chnames['\u00c5'] = _("A RING")

# Translators: this is the spoken word for the character 'Æ' (U+00c6)
#
chnames['\u00c6'] = _("A E")

# Translators: this is the spoken word for the character 'Ç' (U+00c7)
#
chnames['\u00c7'] = _("C CEDILLA")

# Translators: this is the spoken word for the character 'È' (U+00c8)
#
chnames['\u00c8'] = _("E GRAVE")

# Translators: this is the spoken word for the character 'É' (U+00c9)
#
chnames['\u00c9'] = _("E ACUTE")

# Translators: this is the spoken word for the character 'Ê' (U+00ca)
#
chnames['\u00ca'] = _("E CIRCUMFLEX")

# Translators: this is the spoken word for the character 'Ë' (U+00cb)
#
chnames['\u00cb'] = _("E UMLAUT")

# Translators: this is the spoken word for the character 'Ì' (U+00cc)
#
chnames['\u00cc'] = _("I GRAVE")

# Translators: this is the spoken word for the character 'Í' (U+00cd)
#
chnames['\u00cd'] = _("I ACUTE")

# Translators: this is the spoken word for the character 'Î' (U+00ce)
#
chnames['\u00ce'] = _("I CIRCUMFLEX")

# Translators: this is the spoken word for the character 'Ï' (U+00cf)
#
chnames['\u00cf'] = _("I UMLAUT")

# Translators: this is the spoken word for the character 'Ð' (U+00d0)
#
chnames['\u00d0'] = _("ETH")

# Translators: this is the spoken word for the character 'Ñ' (U+00d1)
#
chnames['\u00d1'] = _("N TILDE")

# Translators: this is the spoken word for the character 'Ò' (U+00d2)
#
chnames['\u00d2'] = _("O GRAVE")

# Translators: this is the spoken word for the character 'Ó' (U+00d3)
#
chnames['\u00d3'] = _("O ACUTE")

# Translators: this is the spoken word for the character 'Ô' (U+00d4)
#
chnames['\u00d4'] = _("O CIRCUMFLEX")

# Translators: this is the spoken word for the character 'Õ' (U+00d5)
#
chnames['\u00d5'] = _("O TILDE")

# Translators: this is the spoken word for the character 'Ö' (U+00d6)
#
chnames['\u00d6'] = _("O UMLAUT")

# Translators: this is the spoken word for the character '×' (U+00d7)
#
chnames['\u00d7'] = _("times")

# Translators: this is the spoken word for the character 'Ø' (U+00d8)
#
chnames['\u00d8'] = _("O STROKE")

# Translators: this is the spoken word for the character 'Ù' (U+00d9)
#
chnames['\u00d9'] = _("U GRAVE")

# Translators: this is the spoken word for the character 'Ú' (U+00da)
#
chnames['\u00da'] = _("U ACUTE")

# Translators: this is the spoken word for the character 'Û' (U+00db)
#
chnames['\u00db'] = _("U CIRCUMFLEX")

# Translators: this is the spoken word for the character 'Ü' (U+00dc)
#
chnames['\u00dc'] = _("U UMLAUT")

# Translators: this is the spoken word for the character 'Ý' (U+00dd)
#
chnames['\u00dd'] = _("Y ACUTE")

# Translators: this is the spoken word for the character 'Þ' (U+00de)
#
chnames['\u00de'] = _("THORN")

# Translators: this is the spoken word for the character 'ß' (U+00df)
#
chnames['\u00df'] = _("s sharp")

# Translators: this is the spoken word for the character 'à' (U+00e0)
#
chnames['\u00e0'] = _("a grave")

# Translators: this is the spoken word for the character 'â' (U+00e2)
#
chnames['\u00e2'] = _("a circumflex")

# Translators: this is the spoken word for the character 'ã' (U+00e3)
#
chnames['\u00e3'] = _("a tilde")

# Translators: this is the spoken word for the character 'ä' (U+00e4)
#
chnames['\u00e4'] = _("a umlaut")

# Translators: this is the spoken word for the character 'å' (U+00e5)
#
chnames['\u00e5'] = _("a ring")

# Translators: this is the spoken word for the character 'æ' (U+00e6)
#
chnames['\u00e6'] = _("a e")

# Translators: this is the spoken word for the character 'ç' (U+00e7)
#
chnames['\u00e7'] = _("c cedilla")

# Translators: this is the spoken word for the character 'è' (U+00e8)
#
chnames['\u00e8'] = _("e grave")

# Translators: this is the spoken word for the character 'é' (U+00e9)
#
chnames['\u00e9'] = _("e acute")

# Translators: this is the spoken word for the character 'ê' (U+00ea)
#
chnames['\u00ea'] = _("e circumflex")

# Translators: this is the spoken word for the character 'ë' (U+00eb)
#
chnames['\u00eb'] = _("e umlaut")

# Translators: this is the spoken word for the character 'ì' (U+00ec)
#
chnames['\u00ec'] = _("i grave")

# Translators: this is the spoken word for the character 'í' (U+00ed)
#
chnames['\u00ed'] = _("i acute")

# Translators: this is the spoken word for the character 'î' (U+00ee)
#
chnames['\u00ee'] = _("i circumflex")

# Translators: this is the spoken word for the character 'ï' (U+00ef)
#
chnames['\u00ef'] = _("i umlaut")

# Translators: this is the spoken word for the character 'ð' (U+00f0)
#
chnames['\u00f0'] = _("eth")

# Translators: this is the spoken word for the character 'ñ' (U+00f1)
#
chnames['\u00f1'] = _("n tilde")

# Translators: this is the spoken word for the character 'ò' (U+00f2)
#
chnames['\u00f2'] = _("o grave")

# Translators: this is the spoken word for the character 'ó' (U+00f3)
#
chnames['\u00f3'] = _("o acute")

# Translators: this is the spoken word for the character 'ô' (U+00f4)
#
chnames['\u00f4'] = _("o circumflex")

# Translators: this is the spoken word for the character 'õ' (U+00f5)
#
chnames['\u00f5'] = _("o tilde")

# Translators: this is the spoken word for the character 'ö' (U+00f6)
#
chnames['\u00f6'] = _("o umlaut")

# Translators: this is the spoken word for the character '÷' (U+00f7)
#
chnames['\u00f7'] = _("divided by")

# Translators: this is the spoken word for the character 'ø' (U+00f8)
#
chnames['\u00f8'] = _("o stroke")

# Translators: this is the spoken word for the character 'þ' (U+00fe)
#
chnames['\u00fe'] = _("thorn")

# Translators: this is the spoken word for the character 'ú' (U+00fa)
#
chnames['\u00fa'] = _("u acute")

# Translators: this is the spoken word for the character 'ù' (U+00f9)
#
chnames['\u00f9'] = _("u grave")

# Translators: this is the spoken word for the character 'û' (U+00fb)
#
chnames['\u00fb'] = _("u circumflex")

# Translators: this is the spoken word for the character 'ü' (U+00fc)
#
chnames['\u00fc'] = _("u umlaut")

# Translators: this is the spoken word for the character 'ý' (U+00fd)
#
chnames['\u00fd'] = _("y acute")

# Translators: this is the spoken word for the character 'ÿ' (U+00ff)
#
chnames['\u00ff'] = _("y umlaut")

# Translators: this is the spoken word for the character 'Ÿ' (U+0178)
#
chnames['\u0178'] = _("Y UMLAUT")

# Translators: this is the spoken word for the character 'ƒ' (U+0192)
#
chnames['\u0192'] = _("florin")

# Translators: this is the spoken word for the character '–' (U+2013)
#
chnames['\u2013'] = _("en dash")

# Translators: this is the spoken word for the left single quote: ‘
# (U+2018)
#
chnames['\u2018'] = _("left single quote")

# Translators: this is the spoken word for the right single quote: ’
# (U+2019)
#
chnames['\u2019'] = _("right single quote")

# Translators: this is the spoken word for the character '‚' (U+201a)
#
chnames['\u201a'] = _("single low quote")

# Translators: this is the spoken word for the character '“' (U+201c)
#
chnames['\u201c'] = _("left double quote")

# Translators: this is the spoken word for the character '”' (U+201d)
#
chnames['\u201d'] = _("right double quote")

# Translators: this is the spoken word for the character '„' (U+201e)
#
chnames['\u201e'] = _("double low quote")

# Translators: this is the spoken word for the character '†' (U+2020)
#
chnames['\u2020'] = _("dagger")

# Translators: this is the spoken word for the character '‡' (U+2021)
#
chnames['\u2021'] = _("double dagger")

# Translators: this is the spoken word for the character '•' (U+2022)
#
chnames['\u2022'] = _("bullet")

# Translators: this is the spoken word for the character '‣' (U+2023)
#
chnames['\u2023'] = _("triangular bullet")

# Translators: this is the spoken word for the character '‰' (U+2030)
#
chnames['\u2030'] = _("per mille")

# Translators: this is the spoken word for the character '′' (U+2032)
#
chnames['\u2032'] = _("prime")

# Translators: this is the spoken word for the character '″' (U+2033)
#
chnames['\u2033'] = _("double prime")

# Translators: this is the spoken word for the character '⁃' (U+2043)
#
chnames['\u2043'] = _("hyphen bullet")

# Translators: this is the spoken word for the character '€' (U+20ac)
#
chnames['\u20ac'] = _("euro")

# Translators: this is the spoken word for the character '™' (U+2122)
#
chnames['\u2122'] = _("trademark")

# Translators: this is the spoken word for the character '←' (U+2190)
#
chnames['\u2190'] = _("left arrow")

# Translators: this is the spoken word for the character '→' (U+2192)
#
chnames['\u2192'] = _("right arrow")

# Translators: this is the spoken word for the character '≈' (U+2248)
#
chnames['\u2248'] = _("almost equal to")

# Translators: this is the spoken word for the character '≠' (U+2260)
#
chnames['\u2260'] = _("not equal to")

# Translators: this is the spoken word for the character '≤' (U+2264)
#
chnames['\u2264'] = _("less than or equal to")

# Translators: this is the spoken word for the character '≥' (U+2265)
#
chnames['\u2265'] = _("greater than or equal to")

# Translators: this is the spoken word for the character '√' (U+221a)
#
chnames['\u221a'] = _("square root")

# Translators: this is the spoken word for the character '∛' (U+221b)
#
chnames['\u221b'] = _("cube root")

# Translators: this is the spoken word for the character '∞' (U+221e)
#
chnames['\u221e'] = _("infinity")

# Translators: this is the spoken word for the character '■' (U+25a0)
# It can be used as a bullet in a list.
#
chnames['\u25a0'] = _("black square")

# Translators: this is the spoken word for the character '□' (U+25a1)
# It can be used as a bullet in a list.
#
chnames['\u25a1'] = _("white square")

# Translators: this is the spoken word for the character '◆' (U+25c6)
# It can be used as a bullet in a list.
#
chnames['\u25c6'] = _("black diamond")

# Translators: this is the spoken word for the character '○' (U+25cb)
# It can be used as a bullet in a list.
#
chnames['\u25cb'] = _("white circle")

# Translators: this is the spoken word for the character '●' (U+25cf)
# It can be used as a bullet in a list.
#
chnames['\u25cf'] = _("black circle")

# Translators: this is the spoken word for the character '◦' (U+25e6)
#
chnames['\u25e6'] = _("white bullet")

# Translators: this is the spoken word for the character '✓' (U+2713)
# It can be used as a bullet in a list.
#
chnames['\u2713'] = _("check mark")

# Translators: this is the spoken word for the character '✔' (U+2714)
# It can be used as a bullet in a list.
#
chnames['\u2714'] = _("heavy check mark")

# Translators: this is the spoken word for the character 'x' (U+2717)
# This symbol is included here because it can be used as a bullet in 
# an OOo list.  The goal is to inform the user of the appearance of 
# the bullet, while making it clear that it is a bullet and not simply 
# the typed letter 'x'.  "Ballot x" might confuse the user.  Hence the 
# use of "x-shaped bullet".
#
chnames['\u2717'] = _("x-shaped bullet")

# Translators: this is the spoken word for the character '➔' (U+2794)
# This symbol is included here because it can be used as a bullet in 
# an OOo list. The goal is to inform the user of the appearance of 
# the bullet without too much verbiage, hence simply "right-pointing arrow".
#
chnames['\u2794'] = _("right-pointing arrow")

# Translators: this is the spoken word for the character '⁰' (U+2070)
#
chnames['\u2070'] = _("superscript 0")

# Translators: this is the spoken word for the character '⁴' (U+2074)
#
chnames['\u2074'] = _("superscript 4")

# Translators: this is the spoken word for the character '⁵' (U+2075)
#
chnames['\u2075'] = _("superscript 5")

# Translators: this is the spoken word for the character '⁶' (U+2076)
#
chnames['\u2076'] = _("superscript 6")

# Translators: this is the spoken word for the character '⁷' (U+2077)
#
chnames['\u2077'] = _("superscript 7")

# Translators: this is the spoken word for the character '⁸' (U+2078)
#
chnames['\u2078'] = _("superscript 8")

# Translators: this is the spoken word for the character '⁹' (U+2079)
#
chnames['\u2079'] = _("superscript 9")

# Translators: this is the spoken word for the character '⁺' (U+207a)
#
chnames['\u207a'] = _("superscript plus")

# Translators: this is the spoken word for the character '⁻' (U+207b)
#
chnames['\u207b'] = _("superscript minus")

# Translators: this is the spoken word for the character '⁼' (U+207c)
#
chnames['\u207c'] = _("superscript equals")

# Translators: this is the spoken word for the character '⁽' (U+207d)
#
chnames['\u207d'] = _("superscript left paren")

# Translators: this is the spoken word for the character '⁾' (U+207e)
#
chnames['\u207e'] = _("superscript right paren")

# Translators: this is the spoken word for the character 'ⁿ' (U+207f)
#
chnames['\u207f'] = _("superscript n")

# Translators: this is the spoken word for the character '₀' (U+2080)
#
chnames['\u2080'] = _("subscript 0")

# Translators: this is the spoken word for the character '₁' (U+2081)
#
chnames['\u2081'] = _("subscript 1")

# Translators: this is the spoken word for the character '₂' (U+2082)
#
chnames['\u2082'] = _("subscript 2")

# Translators: this is the spoken word for the character '₃' (U+2083)
#
chnames['\u2083'] = _("subscript 3")

# Translators: this is the spoken word for the character '₄' (U+2084)
#
chnames['\u2084'] = _("subscript 4")

# Translators: this is the spoken word for the character '₅' (U+2085)
#
chnames['\u2085'] = _("subscript 5")

# Translators: this is the spoken word for the character '₆' (U+2086)
#
chnames['\u2086'] = _("subscript 6")

# Translators: this is the spoken word for the character '₇' (U+2087)
#
chnames['\u2087'] = _("subscript 7")

# Translators: this is the spoken word for the character '₈' (U+2088)
#
chnames['\u2088'] = _("subscript 8")

# Translators: this is the spoken word for the character '₉' (U+2089)
#
chnames['\u2089'] = _("subscript 9")

# Translators: this is the spoken word for the character '₊' (U+208a)
#
chnames['\u208a'] = _("subscript plus")

# Translators: this is the spoken word for the character '₋' (U+208b)
#
chnames['\u208b'] = _("subscript minus")

# Translators: this is the spoken word for the character '₌' (U+208c)
#
chnames['\u208c'] = _("subscript equals")

# Translators: this is the spoken word for the character '₍' (U+208d)
#
chnames['\u208d'] = _("subscript left paren")

# Translators: this is the spoken word for the character '₎' (U+208e)
#
chnames['\u208e'] = _("subscript right paren")

# Translators: this is the spoken word for the character '➢' (U+27a2)
# This symbol is included here because it can be used as a bullet in an 
# OOo list. The goal is to inform the user of the appearance of the bullet 
# without too much verbiage, hence simply "right-pointing arrowhead".
#
chnames['\u27a2'] = _("right-pointing arrowhead")

# Translators:  StarOffice/OOo includes private-use unicode character U+E00A 
# as a bullet which looks like the black square: ■ (U+25A0).  Therefore, 
# please use the same translation for this character.
#
chnames['\ue00a'] = _("black square")

# Translators:  StarOffice/OOo includes private-use unicode character U+E00C 
# as a bullet which looks like the black diamond: ◆ (U+25C6).  Therefore, 
# please use the same translation for this character.
#
chnames['\ue00c'] = _("black diamond")

def getCharacterName(character):
    """Given a character, return its name as people might refer to it
    in ordinary conversation.

    Arguments:
    - character: the character to get the name for

    Returns a string representing the name for the character
    """

    return chnames.get(character, character)



# Orca
#
# Copyright (C) 2010-2012 Igalia, S.L.
#
# Author: Alejandro Pinheiro Iglesias <apinheiro@igalia.com>
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2010-2012 Igalia, S.L."
__license__   = "LGPL"

import pyatspi
import time
from gi.repository import Gdk

import orca.orca as orca
import orca.scripts.default as default
import orca.debug as debug

# Set with non printable unicode categories. Full table:
# http://www.fileformat.info/info/unicode/category/index.htm
#
non_printable_set = ('Cc', 'Cf', 'Cn', 'Co', 'Cs')

########################################################################
#                                                                      #
# Utility methods.                                                     #
#                                                                      #
########################################################################

def _unicharIsPrint(unichar):
    """ Checks if the unichar is printable

    Equivalent to g_unichar_isprint

    Arguments:
    - unichar: unichar to check if it is printable
    """
    try:
        import unicodedata
        category = unicodedata.category (unichar)
        result = category not in non_printable_set
    except:
        # Normally a exception is because there are a string
        # instead of a single unicode, 'Control_L'
        #
        result = False

    return result

def _computeIsText(string):
    """Decides if the string representation of a keyboard event is
    text or not

    Based on the at-spi equivalent code.

    Arguments:
    - string: a string representation of a keyboardEvent.
    """
    is_text = False

    if string:
        if _unicharIsPrint(string):
            is_text = True
        else:
            is_text = False

    return is_text

def _parentDialog(obj):
    """Looks for an object of ROLE_DIALOG in the ancestry of obj.

    Arguments:
    - obj: an accessible object

    Returns the accessible object if found; otherwise None.
    """

    isDialog = lambda x: x and x.getRole() == pyatspi.ROLE_DIALOG

    return pyatspi.utils.findAncestor(obj, isDialog)

########################################################################
#                                                                      #
# The Cally script class.                                              #
#                                                                      #
########################################################################

class Script(default.Script):

    def __init__(self, app):
        """Creates a new script for Cally applications.

        Arguments:
        - app: the application to create a script for.
        """

        default.Script.__init__(self, app)
        self._activeDialog = (None, 0) # (Accessible, Timestamp)
        self._activeDialogLabels = {}  # key == hash(obj), value == name

    def checkKeyboardEventData(self, keyboardEvent):
        """Processes the given keyboard event.

        Here is used to:
        * Fill event_string using the key.id
        * Set the is_text properly

        Arguments:
        - keyboardEvent: an instance of input_event.KeyboardEvent
        """

        # On the AtkKeyEventStruct documentation you can find this
        # description:
        # guint keyval;
        # A guint representing a keysym value corresponding to those
        # used by GDK
        #
        # There are no Clutter way to get a gdk-like keyvalname.
        # Anyway, cally will fill event_string with the final
        # representation of a text char.
        #
        # In the same way, Clutter provides the keyval without the
        # modifiers, and GDK yes. We will try to apply it, at least
        # to compute keyval_name
        #
        # More information:
        # http://library.gnome.org/devel/atk/stable/AtkUtil.html
        # http://bugzilla.o-hand.com/show_bug.cgi?id=2072

        # apply the modifiers to keyboardEvent.id
        #
        keyval = keyboardEvent.id
        try:
            keymap = Gdk.Keymap.get_default()

            if keymap:
                success, entries = keymap.get_entries_for_keyval(keyval)
                group = entries[0].group
                modifiers = Gdk.ModifierType(keyboardEvent.modifiers)
                success, keyval, egroup, level, consumed = \
                    keymap.translate_keyboard_state (keyboardEvent.hw_code,
                                                     modifiers,
                                                     group)
        except:
            debug.println(debug.LEVEL_FINE,
                          "Could not compute keyval with modifiers")

        string = "prev keyval=%d" % keyboardEvent.id
        string = string + " post keyval=%d" % keyval

        debug.println(debug.LEVEL_FINE, string)

        keyboardEvent.id = keyval

        # if cally doesn't provide a event_string we get that using
        # Gdk. I know that it will probably called again computing
        # keyval_name but to simplify code, and not start to add
        # guess-code here I will maintain that in this way
        #
        if (keyboardEvent.event_string == ""):
            debug.println (debug.LEVEL_FINE, "Computing event_string")
            try:
                keyboardEvent.event_string = Gdk.keyval_name(keyboardEvent.id)
            except:
                debug.println(debug.LEVEL_FINE,
                              "Could not obtain keyval_name for id: %d" \
                                  % keyboardEvent.id)

            # at-spi uses event_string to compute is_text, so if it is
            # NULL we should compute again with the proper
            # event_string
            #
            keyboardEvent.is_text = _computeIsText(keyboardEvent.event_string)

        return default.Script.checkKeyboardEventData(self, keyboardEvent)

    def skipObjectEvent(self, event):
        """Gives us, and scripts, the ability to decide an event isn't
        worth taking the time to process under the current circumstances.

        Arguments:
        - event: the Event

        Returns True if we shouldn't bother processing this object event.
        """

        try:
            role = event.source.getRole()
        except:
            pass
        else:
            # We must handle all dialogs ourselves in this script.
            if role == pyatspi.ROLE_DIALOG:
                return False

        return default.Script.skipObjectEvent(self, event)

    def presentDialogLabel(self, event):
        """Examines and, if appropriate, presents a new or changed label
        found in a dialog box. Returns True if we handled the presentation
        here."""

        try:
            role = event.source.getRole()
            name = event.source.name
        except:
            return False

        activeDialog, timestamp = self._activeDialog
        if not activeDialog or role != pyatspi.ROLE_LABEL:
            return False

        obj = hash(event.source)
        if name == self._activeDialogLabels.get(obj):
            return True

        if activeDialog == _parentDialog(event.source):
            self.presentMessage(name)
            self._activeDialogLabels[obj] = name
            return True

        return False

    def onNameChanged(self, event):
        """Called whenever a property on an object changes.

        Arguments:
        - event: the Event
        """

        if self.presentDialogLabel(event):
            return

        default.Script.onNameChanged(self, event)

    def onStateChanged(self, event):
        """Called whenever an object's state changes.

        Arguments:
        - event: the Event
        """

        try:
            role = event.source.getRole()
            name = event.source.name
            state = event.source.getState()
        except:
            return

        activeDialog, timestamp = self._activeDialog
        eType = event.type
        if eType.startswith("object:state-changed:showing"):
            # When entering overview with many open windows, we get quite
            # a few state-changed:showing events for nameless panels. The
            # act of processing these by the default script causes us to
            # present nothing, and introduces a significant delay before
            # presenting the Top Bar button when Ctrl+Alt+Tab was pressed.
            if role == pyatspi.ROLE_PANEL and not name:
                return

            # We cannot count on events or their order from dialog boxes.
            # Therefore, the only way to reliably present a dialog is by
            # ignoring the events of the dialog itself and keeping track
            # of the current dialog.
            if not event.detail1 and event.source == activeDialog:
                self._activeDialog = (None, 0)
                self._activeDialogLabels = {}
                return

            if activeDialog and role == pyatspi.ROLE_LABEL and event.detail1:
                if self.presentDialogLabel(event):
                    return

        elif eType.startswith("object:state-changed:focused") and event.detail1:
            # The dialog will get presented when its first child gets focus.
            if role == pyatspi.ROLE_DIALOG:
                return

            # This is to present dialog boxes which are, to the user, newly
            # activated. And if something is claiming to be focused that is
            # not in a dialog, that's good to know as well, so update our
            # state regardless.
            if not activeDialog:
                dialog = _parentDialog(event.source)
                self._activeDialog = (dialog, time.time())
                if dialog:
                    orca.setLocusOfFocus(None, dialog)
                    labels = self.utilities.unrelatedLabels(dialog)
                    for label in labels:
                        self._activeDialogLabels[hash(label)] = label.name

        elif eType.startswith("object:state-changed:selected") and event.detail1:
            # Some buttons, like the Wikipedia button, claim to be selected but
            # lack STATE_SELECTED. The other buttons, such as in the Dash and
            # event switcher, seem to have the right state. Since the ones with
            # the wrong state seem to be things we don't want to present anyway
            # we'll stop doing so and hope we are right.
            if state.contains(pyatspi.STATE_SELECTED):
                orca.setLocusOfFocus(event, event.source)
                return

        default.Script.onStateChanged(self, event)

    def getTextLineAtCaret(self, obj, offset=None):
        """Gets the line of text where the caret is."""

        # TODO - JD/API: This is to work around the braille issue reported
        # in bgo 677221. When that is resolved, this workaround can be
        # removed.
        string, caretOffset, startOffset = \
            default.Script.getTextLineAtCaret(self, obj, offset)

        if string:
            return [string, caretOffset, startOffset]

        try:
            text = obj.queryText()
        except:
            pass
        else:
            string = text.getText(0, -1)

        return [string, caretOffset, startOffset]



# Orca
#
# Copyright 2004-2009 Sun Microsystems Inc.
# Copyright 2010-2011 The Orca Team
# Copyright 2012 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""The main module for the Orca screen reader."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2004-2009 Sun Microsystems Inc." \
                "Copyright (c) 2010-2011 The Orca Team" \
                "Copyright (c) 2012 Igalia, S.L."
__license__   = "LGPL"

import importlib
import os
import pyatspi
import re
import signal
import subprocess
import sys

try:
    from gi.repository.Gio import Settings
    a11yAppSettings = Settings('org.gnome.desktop.a11y.applications')
except:
    a11yAppSettings = None

try:
    # This can fail due to gtk not being available.  We want to
    # be able to recover from that if possible.  The main driver
    # for this is to allow "orca --text-setup" to work even if
    # the desktop is not running.
    #
    from gi.repository import Gtk
    from gi.repository import Gdk
    # Note: This last import is here due to bgo #673396.
    # See bgo#673397 for the rest of the story.
    from gi.repository.GdkX11 import X11Screen
except:
    pass

from . import braille
from . import debug
from . import event_manager
from . import messages
from . import notification_messages
from . import orca_state
from . import script_manager
from . import settings
from . import settings_manager
from . import speech
from .input_event import BrailleEvent
from .input_event import KeyboardEvent

_eventManager = event_manager.getManager()
_scriptManager = script_manager.getManager()
_settingsManager = settings_manager.getManager()

try:
    # If we don't have an active desktop, we will get a RuntimeError.
    from . import mouse_review
except RuntimeError:
    pass

def onEnabledChanged(gsetting, key):
    try:
        enabled = gsetting.get_boolean(key)
    except:
        return

    if key == 'screen-reader-enabled' and not enabled:
        shutdown()

def getSettingsManager():
    return _settingsManager

EXIT_CODE_HANG = 50

# The user-settings module (see loadUserSettings).
#
_userSettings = None

# A subset of the original Xmodmap info prior to our stomping on it.
# Right now, this is just for the user's chosen Orca modifier(s).
#
_originalXmodmap = ""
_orcaModifiers = settings.DESKTOP_MODIFIER_KEYS + settings.LAPTOP_MODIFIER_KEYS
_capsLockCleared = False
_restoreOrcaKeys = False

########################################################################
#                                                                      #
# METHODS TO HANDLE APPLICATION LIST AND FOCUSED OBJECTS               #
#                                                                      #
########################################################################

def setLocusOfFocus(event, obj, notifyScript=True, force=False):
    """Sets the locus of focus (i.e., the object with visual focus) and
    notifies the script of the change should the script wish to present
    the change to the user.

    Arguments:
    - event: if not None, the Event that caused this to happen
    - obj: the Accessible with the new locus of focus.
    - notifyScript: if True, propagate this event
    - force: if True, don't worry if this is the same object as the
      current locusOfFocus
    """

    if not force and obj == orca_state.locusOfFocus:
        return

    # If this event is not for the currently active script, then just return.
    #
    if event and event.source and \
       event.host_application and orca_state.activeScript:
        currentApp = orca_state.activeScript.app
        try:
            appList = [event.host_application, event.source.getApplication()]
        except (LookupError, RuntimeError):
            appList = []
            debug.println(debug.LEVEL_SEVERE,
                           "orca.setLocusOfFocus() application Error")
        if not currentApp in appList:
            return

    oldLocusOfFocus = orca_state.locusOfFocus
    try:
        # Just to see if we have a valid object.
        oldLocusOfFocus.getRole()
    except:
        # Either it's None or it's an invalid remote object.
        oldLocusOfFocus = None

    orca_state.locusOfFocus = obj
    try:
        app = orca_state.locusOfFocus.getApplication()
    except:
        orca_state.locusOfFocus = None
        if event:
            debug.println(debug.LEVEL_FINE,
                          "LOCUS OF FOCUS: None event='%s'" % event.type)
        else:
            debug.println(debug.LEVEL_FINE,
                          "LOCUS OF FOCUS: None event=None")
    else:
        try:
            appname = "'" + app.name + "'"
        except:
            appname = "None"
        debug.println(debug.LEVEL_FINE,
                      "LOCUS OF FOCUS: app=%s name='%s' role='%s'" \
                      % (appname,
                         orca_state.locusOfFocus.name,
                         orca_state.locusOfFocus.getRoleName()))

        if event:
            debug.println(debug.LEVEL_FINE,
                          "                event='%s'" % event.type)
        else:
            debug.println(debug.LEVEL_FINE,
                          "                event=None")

    if notifyScript and orca_state.activeScript:
        orca_state.activeScript.locusOfFocusChanged(
            event, oldLocusOfFocus, orca_state.locusOfFocus)

########################################################################
#                                                                      #
# METHODS FOR PRE-PROCESSING AND MASSAGING KEYBOARD EVENTS.            #
#                                                                      #
########################################################################

_orcaModifierPressed = False

def _processKeyboardEvent(event):
    """The primary key event handler for Orca.  Keeps track of various
    attributes, such as the lastInputEvent.  Also does key echo as well
    as any local keybindings before passing the event on to the active
    script.  This method is called synchronously from the AT-SPI registry
    and should be performant.  In addition, it must return True if it has
    consumed the event (and False if not).

    Arguments:
    - event: an AT-SPI DeviceEvent

    Returns True if the event should be consumed.
    """
    global _orcaModifierPressed

    # Weed out duplicate and otherwise bogus events.
    keyboardEvent = KeyboardEvent(event)
    debug.println(debug.LEVEL_FINE, keyboardEvent.toString())
    if keyboardEvent.ignoreDueToTimestamp():
        debug.println(debug.LEVEL_FINE, "IGNORING EVENT DUE TO TIMESTAMP")
        return

    # Figure out what we've got.
    isOrcaModifier = keyboardEvent.isOrcaModifier()
    isPressedEvent = keyboardEvent.isPressedKey()
    if isOrcaModifier:
        _orcaModifierPressed = isPressedEvent
    if _orcaModifierPressed:
        keyboardEvent.modifiers |= settings.ORCA_MODIFIER_MASK

    # Update our state.
    orca_state.lastInputEventTimestamp = event.timestamp
    orca_state.lastInputEvent = keyboardEvent
    if not keyboardEvent.isModifierKey():
        keyboardEvent.setClickCount()
        orca_state.lastNonModifierKeyEvent = keyboardEvent

    # Echo it based on what it is and the user's settings.
    script = orca_state.activeScript
    if script:
        if isPressedEvent:
            script.presentationInterrupt()
        keyboardEvent.present()
        if keyboardEvent.isModifierKey() and not isOrcaModifier:
            return False
 
    # Special modes.
    if not isPressedEvent and keyboardEvent.event_string == "Escape":
        script.exitLearnMode(keyboardEvent)
    if orca_state.learnModeEnabled and not keyboardEvent.modifiers:
        if keyboardEvent.event_string == "F1":
            orca_state.learnModeEnabled = False
            return helpForOrca()
        if keyboardEvent.event_string in ["1", "2"]:
            orca_state.learnModeEnabled = False
            return script.listOrcaShortcuts(keyboardEvent)
    if orca_state.capturingKeys:
        return False
    if notification_messages.listNotificationMessagesModeEnabled:
        return notification_messages.listNotificationMessages(keyboardEvent)

    # See if the event manager wants it (i.e. it is bound to a command.
    if _eventManager.processKeyboardEvent(keyboardEvent):
        return True

    # Do any needed xmodmap crap.
    global _restoreOrcaKeys
    if not isPressedEvent:
        if keyboardEvent.event_string in settings.orcaModifierKeys \
           and orca_state.bypassNextCommand:
            _restoreXmodmap()
            _restoreOrcaKeys = True
        elif _restoreOrcaKeys and not orca_state.bypassNextCommand:
            _createOrcaXmodmap()
            _restoreOrcaKeys = False
    elif not keyboardEvent.isModifierKey():
        _orcaModifierPressed = False
        orca_state.bypassNextCommand = False
 
    return isOrcaModifier or orca_state.learnModeEnabled

########################################################################
#                                                                      #
# METHODS FOR PRE-PROCESSING AND MASSAGING BRAILLE EVENTS.             #
#                                                                      #
########################################################################

def _processBrailleEvent(event):
    """Called whenever a  key is pressed on the Braille display.

    Arguments:
    - command: the BrlAPI event for the key that was pressed.

    Returns True if the event was consumed; otherwise False
    """

    consumed = False

    # Braille key presses always interrupt speech.
    #
    event = BrailleEvent(event)
    if event.event['command'] not in braille.dontInteruptSpeechKeys:
        speech.stop()
    orca_state.lastInputEvent = event

    try:
        consumed = _eventManager.processBrailleEvent(event)
    except:
        debug.printException(debug.LEVEL_SEVERE)

    if (not consumed) and orca_state.learnModeEnabled:
        consumed = True

    return consumed

########################################################################
#                                                                      #
# METHODS FOR HANDLING INITIALIZATION, SHUTDOWN, AND USE.              #
#                                                                      #
########################################################################

def _setXmodmap(xkbmap):
    """Set the keyboard map using xkbcomp."""
    p = subprocess.Popen(['xkbcomp', '-w0', '-', os.environ['DISPLAY']],
        stdin=subprocess.PIPE, stdout=None, stderr=None)
    p.communicate(xkbmap)

def _setCapsLockAsOrcaModifier(enable):
    """Enable or disable use of the caps lock key as an Orca modifier key."""
    interpretCapsLineProg = re.compile(
        r'^\s*interpret\s+Caps[_+]Lock[_+]AnyOfOrNone\s*\(all\)\s*{\s*$', re.I)
    capsModLineProg = re.compile(
        r'^\s*action\s*=\s*SetMods\s*\(\s*modifiers\s*=\s*Lock\s*,\s*clearLocks\s*\)\s*;\s*$', re.I)
    normalCapsLineProg = re.compile(
        r'^\s*action\s*=\s*LockMods\s*\(\s*modifiers\s*=\s*Lock\s*\)\s*;\s*$', re.I)
    normalCapsLine = '        action= LockMods(modifiers=Lock);'
    capsModLine =    '        action= SetMods(modifiers=Lock,clearLocks);'
    lines = _originalXmodmap.decode('UTF-8').split('\n')
    foundCapsInterpretSection = False
    for i in range(len(lines)):
        line = lines[i]
        if not foundCapsInterpretSection:
            if interpretCapsLineProg.match(line):
                foundCapsInterpretSection = True
        else:
            if enable:
                if normalCapsLineProg.match(line):
                    lines[i] = capsModLine
                    _setXmodmap(bytes('\n'.join(lines), 'UTF-8'))
                    return
            else:
                if capsModLineProg.match(line):
                    lines[i] = normalCapsLine
                    _setXmodmap(bytes('\n'.join(lines), 'UTF-8'))
                    return
            if line.find('}'):
                # Failed to find the line we need to change
                return

def _createOrcaXmodmap():
    """Makes an Orca-specific Xmodmap so that the keys behave as we
    need them to do. This is especially the case for the Orca modifier.
    """

    global _capsLockCleared

    cmd = []
    if "Caps_Lock" in settings.orcaModifierKeys:
        _setCapsLockAsOrcaModifier(True)
        _capsLockCleared = True
    elif _capsLockCleared:
        _setCapsLockAsOrcaModifier(False)
        _capsLockCleared = False

def _storeXmodmap(keyList):
    """Save the original xmodmap for the keys in keyList before we alter it.

    Arguments:
    - keyList: A list of named keys to look for.
    """

    global _originalXmodmap
    _originalXmodmap = subprocess.check_output(['xkbcomp', os.environ['DISPLAY'], '-'])

def _restoreXmodmap(keyList=[]):
    """Restore the original xmodmap values for the keys in keyList.

    Arguments:
    - keyList: A list of named keys to look for. An empty list means
      to restore the entire saved xmodmap.
    """

    global _capsLockCleared
    _capsLockCleared = False
    p = subprocess.Popen(['xkbcomp', '-w0', '-', os.environ['DISPLAY']],
        stdin=subprocess.PIPE, stdout=None, stderr=None)
    p.communicate(_originalXmodmap)

def loadUserSettings(script=None, inputEvent=None, skipReloadMessage=False):
    """Loads (and reloads) the user settings module, reinitializing
    things such as speech if necessary.

    Returns True to indicate the input event has been consumed.
    """

    debug.println(debug.LEVEL_FINEST, 'INFO: Loading User Settings')

    global _userSettings

    # Shutdown the output drivers and give them a chance to die.

    speech.shutdown()
    braille.shutdown()

    _scriptManager.deactivate()

    reloaded = False
    if _userSettings:
        _profile = _settingsManager.getSetting('activeProfile')[1]
        try:
            _userSettings = _settingsManager.getGeneralSettings(_profile)
            reloaded = True
        except ImportError:
            debug.printException(debug.LEVEL_FINEST)
        except:
            debug.printException(debug.LEVEL_SEVERE)
    else:
        _profile = _settingsManager.profile
        try:
            _userSettings = _settingsManager.getGeneralSettings(_profile)
        except ImportError:
            debug.printException(debug.LEVEL_FINEST)
        except:
            debug.printException(debug.LEVEL_SEVERE)

    if _settingsManager.getSetting('enableSpeech'):
        try:
            speech.init()
            if reloaded and not skipReloadMessage:
                speech.speak(messages.SETTINGS_RELOADED,
                             settings.voices.get(settings.SYSTEM_VOICE))
            debug.println(debug.LEVEL_CONFIGURATION,
                          "Speech module has been initialized.")
        except:
            debug.printException(debug.LEVEL_SEVERE)
            debug.println(debug.LEVEL_SEVERE,
                          "Could not initialize connection to speech.")
    else:
        debug.println(debug.LEVEL_CONFIGURATION,
                      "Speech module has NOT been initialized.")

    if _settingsManager.getSetting('enableBraille'):
        try:
            braille.init(_processBrailleEvent, settings.tty)
        except:
            debug.printException(debug.LEVEL_WARNING)
            debug.println(debug.LEVEL_WARNING,
                          "Could not initialize connection to braille.")

    # I'm not sure where else this should go. But it doesn't really look
    # right here.
    try:
        mouse_review.mouse_reviewer.toggle(on=settings.enableMouseReview)
    except NameError:
        pass

    global _orcaModifiers
    custom = [k for k in settings.orcaModifierKeys if k not in _orcaModifiers]
    _orcaModifiers += custom
    # Handle the case where a change was made in the Orca Preferences dialog.
    #
    if _originalXmodmap:
        _restoreXmodmap(_orcaModifiers)

    _storeXmodmap(_orcaModifiers)
    _createOrcaXmodmap()

    _scriptManager.activate()
    _eventManager.activate()

    debug.println(debug.LEVEL_FINEST, 'INFO: User Settings Loaded')

    return True

def showAppPreferencesGUI(script=None, inputEvent=None):
    """Displays the user interace to configure the settings for a
    specific applications within Orca and set up those app-specific
    user preferences using a GUI.

    Returns True to indicate the input event has been consumed.
    """

    try:
        module = importlib.import_module('.app_gui_prefs', 'orca')
        module.showPreferencesUI()
    except:
        debug.printException(debug.LEVEL_SEVERE)

    return True

def showPreferencesGUI(script=None, inputEvent=None):
    """Displays the user interace to configure Orca and set up
    user preferences using a GUI.

    Returns True to indicate the input event has been consumed.
    """

    try:
        module = importlib.import_module('.orca_gui_prefs', 'orca')
        module.showPreferencesUI()
    except:
        debug.printException(debug.LEVEL_SEVERE)

    return True

def helpForOrca(script=None, inputEvent=None, page=""):
    """Show Orca Help window (part of the GNOME Access Guide).

    Returns True to indicate the input event has been consumed.
    """
    uri = "help:orca"
    if page:
        uri += "?%s" % page
    Gtk.show_uri(Gdk.Screen.get_default(),
                 uri,
                 Gtk.get_current_event_time())
    return True

def quitOrca(script=None, inputEvent=None):
    """Quit Orca. Check if the user wants to confirm this action.
    If so, show the confirmation GUI otherwise just shutdown.

    Returns True to indicate the input event has been consumed.
    """

    shutdown()

    return True

def showFindGUI(script=None, inputEvent=None):
    """Displays the user interace to perform an Orca Find.

    Returns True to indicate the input event has been consumed.
    """

    try:
        module = importlib.import_module('.orca_gui_find', 'orca')
        module.showFindUI()
    except:
        debug.printException(debug.LEVEL_SEVERE)

# If True, this module has been initialized.
#
_initialized = False

def init(registry):
    """Initialize the orca module, which initializes the speech and braille
    modules.  Also builds up the application list, registers for AT-SPI events,
    and creates scripts for all known applications.

    Returns True if the initialization procedure has run, or False if this
    module has already been initialized.
    """

    debug.println(debug.LEVEL_FINEST, 'INFO: Initializing Orca module')

    global _initialized

    if _initialized and _settingsManager.isScreenReaderServiceEnabled():
        return False

    # Do not hang on initialization if we can help it.
    #
    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.signal(signal.SIGALRM, settings.timeoutCallback)
        signal.alarm(settings.timeoutTime)

    loadUserSettings()
    _eventManager.registerKeystrokeListener(_processKeyboardEvent)

    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.alarm(0)

    _initialized = True
    # In theory, we can do this through dbus. In practice, it fails to
    # work sometimes. Until we know why, we need to leave this as-is
    # so that we respond when gnome-control-center is used to stop Orca.
    if a11yAppSettings:
        a11yAppSettings.connect('changed', onEnabledChanged)

    debug.println(debug.LEVEL_FINEST, 'INFO: Orca module initialized')

    return True

def start(registry):
    """Starts Orca.
    """

    debug.println(debug.LEVEL_FINEST, 'INFO: Starting Orca')

    if not _initialized:
        init(registry)

    # Do not hang on startup if we can help it.
    #
    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.signal(signal.SIGALRM, settings.timeoutCallback)
        signal.alarm(settings.timeoutTime)

    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.alarm(0)

    if settings.cacheValues:
        pyatspi.setCacheLevel(pyatspi.CACHE_PROPERTIES)

    debug.println(debug.LEVEL_FINEST, 'INFO: Orca starting registry')
    registry.start(gil=settings.useGILIdleHandler)

def die(exitCode=1):
    pid = os.getpid()
    if exitCode == EXIT_CODE_HANG:
        # Someting is hung and we wish to abort.
        os.kill(pid, signal.SIGKILL)
        return

    shutdown()
    sys.exit(exitCode)
    if exitCode > 1:
        os.kill(pid, signal.SIGTERM)

def timeout(signum=None, frame=None):
    debug.println(debug.LEVEL_SEVERE,
                  "TIMEOUT: something has hung.  Aborting.")
    debug.printStack(debug.LEVEL_ALL)
    debug.examineProcesses()
    die(EXIT_CODE_HANG)

def shutdown(script=None, inputEvent=None):
    """Exits Orca.  Unregisters any event listeners and cleans up.

    Returns True if the shutdown procedure ran or False if this module
    was never initialized.
    """

    debug.println(debug.LEVEL_FINEST, 'INFO: Shutting down Orca')

    global _initialized

    if not _initialized:
        return False

    # Try to say goodbye, but be defensive if something has hung.
    #
    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.signal(signal.SIGALRM, settings.timeoutCallback)
        signal.alarm(settings.timeoutTime)

    orca_state.activeScript.presentMessage(messages.STOP_ORCA)

    _eventManager.deactivate()
    _scriptManager.deactivate()

    # Shutdown all the other support.
    #
    if settings.enableSpeech:
        speech.shutdown()
    if settings.enableBraille:
        braille.shutdown()

    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.alarm(0)

    _initialized = False
    _restoreXmodmap(_orcaModifiers)

    debug.println(debug.LEVEL_FINEST, 'INFO: Orca stopping registry')
    pyatspi.Registry.stop()

    debug.println(debug.LEVEL_FINEST, 'INFO: Orca shutdown complete')

    return True

exitCount = 0
def shutdownOnSignal(signum, frame):
    global exitCount

    debug.println(debug.LEVEL_ALL,
                  "Shutting down and exiting due to signal = %d" \
                  % signum)

    debug.println(debug.LEVEL_ALL, "Current stack is:")
    debug.printStack(debug.LEVEL_ALL)

    # Well...we'll try to exit nicely, but if we keep getting called,
    # something bad is happening, so just quit.
    #
    if exitCount:
        die(signum)
    else:
        exitCount += 1

    # Try to do a graceful shutdown if we can.
    #
    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.signal(signal.SIGALRM, settings.timeoutCallback)
        signal.alarm(settings.timeoutTime)

    try:
        if _initialized:
            shutdown()
        else:
            # We always want to try to shutdown speech since the
            # speech servers are very persistent about living.
            #
            speech.shutdown()
            shutdown()
        cleanExit = True
    except:
        cleanExit = False

    if settings.timeoutCallback and (settings.timeoutTime > 0):
        signal.alarm(0)

    if not cleanExit:
        die(EXIT_CODE_HANG)

def abortOnSignal(signum, frame):
    debug.println(debug.LEVEL_ALL,
                  "Aborting due to signal = %d" \
                  % signum)
    die(signum)

def main():
    """The main entry point for Orca.  The exit codes for Orca will
    loosely be based on signals, where the exit code will be the
    signal used to terminate Orca (if a signal was used).  Otherwise,
    an exit code of 0 means normal completion and an exit code of 50
    means Orca exited because of a hang."""

    # Method to call when we think something might be hung.
    #
    settings.timeoutCallback = timeout

    # Various signal handlers we want to listen for.
    #
    signal.signal(signal.SIGHUP, shutdownOnSignal)
    signal.signal(signal.SIGINT, shutdownOnSignal)
    signal.signal(signal.SIGTERM, shutdownOnSignal)
    signal.signal(signal.SIGQUIT, shutdownOnSignal)
    signal.signal(signal.SIGSEGV, abortOnSignal)

    if not _settingsManager.isAccessibilityEnabled():
        _settingsManager.setAccessibility(True)

    init(pyatspi.Registry)

    try:
        message = messages.START_ORCA
        if not _settingsManager.getSetting('onlySpeakDisplayedText'):
            speech.speak(message, settings.voices.get(settings.SYSTEM_VOICE))
        braille.displayMessage(message)
    except:
        debug.printException(debug.LEVEL_SEVERE)

    script = orca_state.activeScript
    if script:
        window = script.utilities.activeWindow()
        if window and not orca_state.locusOfFocus:
            setLocusOfFocus(None, window)

    try:
        start(pyatspi.Registry) # waits until we stop the registry
    except:
        die(EXIT_CODE_HANG)
    return 0

if __name__ == "__main__":
    sys.exit(main())



# Orca
#
# Copyright 2006-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script for gnome-search-tool"""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
__license__   = "LGPL"

import orca.debug as debug
import orca.messages as messages
import orca.scripts.default as default

from orca.orca_i18n import _

import pyatspi
import time
from gi.repository import GLib
from gi.repository import GObject
GObject.threads_init()


########################################################################
#                                                                      #
# The gnome-search-tool script class.                                  #
#                                                                      #
########################################################################

class Script(default.Script):

    def __init__(self, app):
        """Creates a new script for the given application.

        Arguments:
        - app: the application to create a script for.
        """

        default.Script.__init__(self, app)

        # Set the debug level for all the methods in this script.
        #
        self.debugLevel = debug.LEVEL_FINEST

        # The table of files found.
        #
        self.fileTable = None

        # Set to true if we are doing a search.
        #
        self.searching = False

        # Time value using by the interval timer.
        #
        self.startTime = None

        # Interval in seconds, between utterances of "Searching" when a
        # search is in progress.
        #
        self.searchInterval = 5.0

    def getListeners(self):
        """Sets up the AT-SPI event listeners for this script.
        """
        listeners = default.Script.getListeners(self)

        listeners["object:state-changed:showing"]           = \
            self.onStateChanged

        return listeners

    def _speakSearching(self):
        """If we are still searching, let the user know. Then start another
        timer to go off again and repeat this process.
        """

        if not self.searching:
            return False

        currentTime = time.time()
        if not self.startTime or \
           (currentTime > (self.startTime + self.searchInterval)):
            self.presentMessage(messages.SEARCHING)
            self.startTime = time.time()

        return True

    def onStateChanged(self, event):
        """Called whenever an object's state changes.

        Arguments:
        - event: the Event
        """

        details = debug.getAccessibleDetails(self.debugLevel, event.source)
        debug.printObjectEvent(self.debugLevel, event, details)

        rolesList = [pyatspi.ROLE_PUSH_BUTTON, \
                    pyatspi.ROLE_FILLER, \
                    pyatspi.ROLE_FILLER, \
                    pyatspi.ROLE_FILLER, \
                    pyatspi.ROLE_FRAME, \
                    pyatspi.ROLE_APPLICATION]
        visible = event.source.getState().contains(pyatspi.STATE_VISIBLE)

        # Check to see if we have just had an "object:state-changed:showing"
        # event for the Stop button. If the name is "Stop", and one of its
        # states is VISIBLE, that means we have started a search. As the
        # search progresses, regularly inform the user of this by speaking
        # "Searching" (assuming the search tool has focus).
        #
        # Translators: the "Stop" string must match what gnome-search-tool
        # is using.  We hate keying off stuff like this, but we're forced
        # to do so in this case.
        #
        if self.utilities.hasMatchingHierarchy(event.source, rolesList) and \
           event.source.name == _("Stop") and visible:
            debug.println(self.debugLevel,
                          "gnome-search-tool.onNameChanged - " \
                          + "search started.")

            self.searching = True

            # If we don't already have a handle to the table containing the
            # list of files found, then get it now.
            #
            if not self.fileTable:
                frame = self.utilities.topLevelObject(event.source)
                allTables = self.utilities.descendantsWithRole(
                    frame, pyatspi.ROLE_TABLE)
                self.fileTable = allTables[0]

            GLib.idle_add(self._speakSearching)

        # Check to see if we have just had an "object:state-changed:showing"
        # event for the Find button. If the name is "Find", and one of its
        # states is VISIBLE and we are currently searching, that means we
        # have just stopped a search. Inform the user that the search is
        # complete and tell them how many files were found.
        #
        # Translators: the "Find" string must match what gnome-search-tool
        # is using.  We hate keying off stuff like this, but we're forced
        # to do so in this case.
        #
        if self.utilities.hasMatchingHierarchy(event.source, rolesList) \
           and event.source.name == _("Find") and visible and self.searching:
            debug.println(self.debugLevel,
                          "gnome-search-tool.onNameChanged - " \
                          + "search completed.")

            self.searching = False
            self.presentMessage(messages.SEARCH_COMPLETE)
            sensitive = self.fileTable.getState().contains( \
                                                pyatspi.STATE_SENSITIVE)
            if sensitive:
                try:
                    fileCount = self.fileTable.queryTable().nRows
                except NotImplementedError:
                    fileCount = 0
                noFilesString = messages.filesFound(fileCount)
                self.presentMessage(noFilesString)
            else:
                self.presentMessage(messages.FILES_NOT_FOUND)

        # Pass the event onto the parent class to be handled in the default way.
        #
        default.Script.onStateChanged(self, event)



from orca.speech_generator import SpeechGenerator as BaseSpeechGenerator
import pyatspi

class SpeechGenerator(BaseSpeechGenerator):
    def _generateImage(self, obj, **args):
        """Returns an array of strings (and possibly voice and audio
        specifications) that represent the image on the the object, if
        it exists.  Otherwise, an empty array is returned.
        """
        result = []
        try:
            image = obj.queryImage()
        except:
            pass
        else:
            args['role'] = pyatspi.ROLE_IMAGE
            result.extend(self.generate(obj, **args))
        return result




# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script for StarOffice and OpenOffice."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2005-2009 Sun Microsystems Inc."
__license__   = "LGPL"

import pyatspi

import orca.braille as braille
import orca.braille_generator as braille_generator
import orca.settings_manager as settings_manager

_settingsManager = settings_manager.getManager()

class BrailleGenerator(braille_generator.BrailleGenerator):

    # pylint: disable-msg=W0142

    def __init__(self, script):
        braille_generator.BrailleGenerator.__init__(self, script)

    def _generateRoleName(self, obj, **args):
        result = []
        role = args.get('role', obj.getRole())
        if role != pyatspi.ROLE_DOCUMENT_FRAME:
            result.extend(braille_generator.BrailleGenerator._generateRoleName(
                self, obj, **args))
        return result

    def _generateRowHeader(self, obj, **args):
        """Returns an array of strings that represent the row header for an
        object that is in a table, if it exists.  Otherwise, an empty
        array is returned. Overridden here so that we can get the
        dynamic row header(s).
        """
        result = []
        try:
            table = obj.parent.queryTable()
        except:
            pass
        else:
            index = self._script.utilities.cellIndex(obj)
            rowIndex = table.getRowAtIndex(index)
            if rowIndex >= 0 \
               and hash(obj.parent) in self._script.dynamicRowHeaders:
                column = self._script.dynamicRowHeaders[hash(obj.parent)]
                header = self._script.getDynamicColumnHeaderCell(obj, column)
                try:
                    headerText = header.queryText()
                except:
                    headerText = None
                if header.childCount > 0:
                    for child in header:
                        text = self._script.utilities.substring(child, 0, -1)
                        if text:
                            result.append(text)
                elif headerText:
                    text = self._script.utilities.substring(header, 0, -1)
                    if text:
                        result.append(text)
        return result

    def _generateColumnHeader(self, obj, **args):
        """Returns an array of strings that represent the column header for an
        object that is in a table, if it exists.  Otherwise, an empty
        array is returned. Overridden here so that we can get the
        dynamic column header(s).
        """
        result = []
        try:
            table = obj.parent.queryTable()
        except:
            pass
        else:
            index = self._script.utilities.cellIndex(obj)
            columnIndex = table.getColumnAtIndex(index)
            if columnIndex >= 0 \
               and hash(obj.parent) in self._script.dynamicColumnHeaders:
                row = self._script.dynamicColumnHeaders[hash(obj.parent)]
                header = self._script.getDynamicRowHeaderCell(obj, row)
                try:
                    headerText = header.queryText()
                except:
                    headerText = None
                if header.childCount > 0:
                    for child in header:
                        text = self._script.utilities.substring(child, 0, -1)
                        if text:
                            result.append(text)
                elif headerText:
                    text = self._script.utilities.substring(header, 0, -1)
                    if text:
                        result.append(text)
        return result

    def _generateSpreadSheetCell(self, obj, **args):
        result = []
        if self._script.inputLineForCell == None:
            self._script.inputLineForCell = \
                self._script.locateInputLine(obj)
        # If the spread sheet table cell has something in it, then we want
        # to append the name of the cell (which will be its location). Note
        # that if the cell was empty, self._script.utilities.displayedText
        # will have already done this for us.
        #
        try:
            if obj.queryText():
                objectText = self._script.utilities.substring(obj, 0, -1)
                if objectText and len(objectText) != 0:
                    result.append(braille.Component(
                        obj, objectText + " " + obj.name))
                else:
                    result.append(braille.Component(obj, obj.name))
        except:
            pass
        return result

    def _generateRealTableCell(self, obj, **args):
        """Get the speech for a table cell. If this isn't inside a
        spread sheet, just return the utterances returned by the default
        table cell speech handler.

        Arguments:
        - obj: the table cell

        Returns a list of utterances to be spoken for the object.
        """
        result = []
        if self._script.isSpreadSheetCell(obj):
            result.extend(self._generateSpreadSheetCell(obj, **args))
        else:
            # Check to see how many children this table cell has. If it's
            # just one (or none), then pass it on to the superclass to be
            # processed.
            #
            # If it's more than one, then get the speech for each child,
            # and call this method again.
            #
            if obj.childCount <= 1:
                result.extend(braille_generator.BrailleGenerator.\
                              _generateRealTableCell(self, obj, **args))
            else:
                for child in obj:
                    cellResult = self._generateRealTableCell(child, **args)
                    if cellResult and result and self._mode == 'braille':
                        delimiter = _settingsManager.getSetting(
                            'brailleTableCellDelimiter')
                        result.append(braille.Region(delimiter))
                    result.extend(cellResult)
        return result

    def _generateTableCellRow(self, obj, **args):
        """Get the speech for a table cell row or a single table cell
        if settings.readTableCellRow is False. If this isn't inside a
        spread sheet, just return the utterances returned by the default
        table cell speech handler.

        Arguments:
        - obj: the table cell

        Returns a list of utterances to be spoken for the object.
        """
        result = []
        if self._script.isSpreadSheetCell(obj):
            # Adding in a check here to make sure that the parent is a
            # valid table. It's possible that the parent could be a
            # table cell too (see bug #351501).
            #
            parent = obj.parent
            parentTable = parent.queryTable()
            if _settingsManager.getSetting('readTableCellRow') and parentTable:
                index = self._script.utilities.cellIndex(obj)
                row = parentTable.getRowAtIndex(index)
                column = parentTable.getColumnAtIndex(index)
                # This is an indication of whether we should present all the
                # table cells (the user has moved focus up or down a row),
                # or just the current one (focus has moved left or right in
                # the same row).
                #
                presentAll = True
                if "lastRow" in self._script.pointOfReference and \
                    "lastColumn" in self._script.pointOfReference:
                    pointOfReference = self._script.pointOfReference
                    presentAll = \
                        (self._mode == 'braille') \
                        or ((pointOfReference["lastRow"] != row) \
                            or ((row == 0 or row == parentTable.nRows-1) \
                                and pointOfReference["lastColumn"] == column))
                if presentAll:
                    [startIndex, endIndex] = \
                        self._script.getSpreadSheetRowRange(obj)
                    for i in range(startIndex, endIndex+1):
                        cell = parentTable.getAccessibleAt(row, i)
                        showing = cell.getState().contains( \
                                      pyatspi.STATE_SHOWING)
                        if showing:
                            cellResult = self._generateRealTableCell(cell,
                                                                     **args)
                            if cellResult and result \
                               and self._mode == 'braille':
                                delimiter = _settingsManager.getSetting(
                                    'brailleTableCellDelimiter')
                                result.append(braille.Region(delimiter))
                            result.extend(cellResult)
                else:
                    result.extend(self._generateRealTableCell(obj, **args))
            else:
                result.extend(self._generateRealTableCell(obj, **args))
        else:
            result.extend(
                braille_generator.BrailleGenerator._generateTableCellRow(
                    self, obj, **args))
        return result

    def _generateChildTab(self, obj, **args):
        """If we are in the slide presentation scroll pane, also announce the
        current page tab. See bug #538056 for more details.
        """
        result = []
        rolesList = [pyatspi.ROLE_SCROLL_PANE, \
                     pyatspi.ROLE_PANEL, \
                     pyatspi.ROLE_PANEL, \
                     pyatspi.ROLE_ROOT_PANE, \
                     pyatspi.ROLE_FRAME, \
                     pyatspi.ROLE_APPLICATION]
        if self._script.utilities.hasMatchingHierarchy(obj, rolesList):
            for child in obj.parent:
                if child.getRole() == pyatspi.ROLE_PAGE_TAB_LIST:
                    for tab in child:
                        eventState = tab.getState()
                        if eventState.contains(pyatspi.STATE_SELECTED):
                            args['role'] = tab.getRole()
                            result.extend(self.generate(tab, **args))
        return result

    def generateBraille(self, obj, **args):
        result = []
        args['useDefaultFormatting'] = \
            ((obj.getRole() == pyatspi.ROLE_LIST) \
                and (not obj.getState().contains(pyatspi.STATE_FOCUSABLE)))
        result.extend(braille_generator.BrailleGenerator.\
                          generateBraille(self, obj, **args))
        del args['useDefaultFormatting']
        return result
